@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  * 
- * Copyright Ericsson AB 2018. All Rights Reserved.
+ * Copyright Ericsson AB 2018-2023. All Rights Reserved.
  * 
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -30,6 +30,7 @@
  *              - Persistent monitor message
  *              - Link
  *              - Unlink
+ *              - Unlink Ack
  *              - Group leader
  *              - Is process alive
  *              - Process info request
@@ -89,6 +90,7 @@
 #endif
 
 struct erl_mesg;
+struct erl_dist_external;
 
 typedef struct {
     struct erl_mesg *next;
@@ -99,6 +101,12 @@ typedef struct {
     Eterm tag;
 } ErtsSignalCommon;
 
+typedef struct {
+    ErtsSignalCommon common;
+    Eterm from;
+    Uint64 id;
+} ErtsSigUnlinkOp;
+
 #define ERTS_SIG_HANDLE_REDS_MAX_PREFERED (CONTEXT_REDS/40)
 
 #ifdef ERTS_PROC_SIG_HARD_DEBUG
@@ -180,6 +188,10 @@ void erts_proc_sig_hdbg_check_in_queue(struct process *c_p, char *what,
 
 struct dist_entry_;
 
+#define ERTS_PROC_HAS_INCOMING_SIGNALS(P)                               \
+    (!!(erts_atomic32_read_nob(&(P)->state)                             \
+        & (ERTS_PSFLG_SIG_Q|ERTS_PSFLG_SIG_IN_Q)))
+
 /*
  * Send operations of currently supported process signals follow...
  */
@@ -210,6 +222,38 @@ void
 erts_proc_sig_send_exit(Process *c_p, Eterm from, Eterm to,
                         Eterm reason, Eterm token, int normal_kills);
 
+/**
+ *
+ * @brief Send an exit signal to a process.
+ *
+ * This function is used instead of erts_proc_sig_send_link_exit()
+ * when the signal arrives via the distribution and
+ * therefore no link structure is available.
+ *
+ * @param[in]     dep           Distribution entry of channel
+ *                              that the signal arrived on.
+ *
+ * @param[in]     from          Identifier of sender.
+ *
+ * @param[in]     to            Identifier of receiver.
+ *
+ * @param[in]     dist_ext      The exit reason in external term format
+ *
+ * @param[in]     hfrag         Heap frag with trace token and dist_ext
+ *                              iff available, otherwise NULL.
+ *
+ * @param[in]     reason        Exit reason.
+ *
+ * @param[in]     token         Seq trace token.
+ *
+ */
+void
+erts_proc_sig_send_dist_exit(DistEntry *dep,
+                             Eterm from, Eterm to,
+                             ErtsDistExternal *dist_ext,
+                             ErlHeapFragment *hfrag,
+                             Eterm reason, Eterm token);
+
 /**
  *
  * @brief Send an exit signal due to broken link to a process.
@@ -260,6 +304,58 @@ erts_proc_sig_send_link_exit(Process *c_p, Eterm from, ErtsLink *lnk,
 int
 erts_proc_sig_send_link(Process *c_p, Eterm to, ErtsLink *lnk);
 
+/**
+ *
+ * @brief Create a new unlink identifier
+ *
+ * The newly created unlink identifier is to be used in an
+ * unlink operation.
+ *
+ * @param[in]     c_p           Pointer to process struct of
+ *                              currently executing process.
+ *
+ * @return                      A new 64-bit unlink identifier
+ *                              unique in context of the
+ *                              calling process. The identifier
+ *                              may be any value but zero.
+ */
+ERTS_GLB_INLINE Uint64 erts_proc_sig_new_unlink_id(Process *c_p);
+
+/**
+ *
+ * @brief Create an unlink op signal structure
+ *
+ * The structure will contain a newly created unlink
+ * identifier to be used in the operation.
+ *
+ * @param[in]     c_p           Pointer to process struct of
+ *                              currently executing process
+ *                              ('from' is a process
+ *                              identifier), or NULL if not
+ *                              called in the context of an
+ *                              executing process ('from' is
+ *                              a port identifier).
+ *
+ * @param[in]     from          Id (as an erlang term) of
+ *                              entity sending the unlink
+ *                              signal.
+ *
+ * @return                      A pointer to the unlink op
+ *                              structure.
+ */
+ErtsSigUnlinkOp *
+erts_proc_sig_make_unlink_op(Process *c_p, Eterm from);
+
+/**
+ *
+ * @brief Destroy an unlink op signal structure
+ *
+ * @param[in]     sulnk         A pointer to the unlink op
+ *                              structure.
+ */
+void
+erts_proc_sig_destroy_unlink_op(ErtsSigUnlinkOp *sulnk);
+
 /**
  *
  * @brief Send an unlink signal to a process.
@@ -268,13 +364,39 @@ erts_proc_sig_send_link(Process *c_p, Eterm to, ErtsLink *lnk);
  * @param[in]     c_p           Pointer to process struct of
  *                              currently executing process.
  *
+ * @param[in]     from          Id (as an erlang term) of
+ *                              entity sending the unlink
+ *                              signal.
+ *
  * @param[in]     lnk           Pointer to link structure from
  *                              the sending side. It should
  *                              contain information about
  *                              receiver.
  */
+Uint64
+erts_proc_sig_send_unlink(Process *c_p, Eterm from, ErtsLink *lnk);
+
+/**
+ *
+ * @brief Send an unlink acknowledgment signal to a process.
+ *
+ *
+ * @param[in]     c_p           Pointer to process struct of
+ *                              currently executing process.
+ *
+ * @param[in]     from          Id (as an erlang term) of
+ *                              entity sending the unlink
+ *                              signal.
+ *
+ * @param[in]     sulnk         A pointer to the unlink op
+ *                              structure. This structure
+ *                              was typically received by
+ *                              the caller in an unlink
+ *                              signal.
+ */
 void
-erts_proc_sig_send_unlink(Process *c_p, ErtsLink *lnk);
+erts_proc_sig_send_unlink_ack(Process *c_p, Eterm from,
+                              ErtsSigUnlinkOp *sulnk);
 
 /**
  *
@@ -282,7 +404,7 @@ erts_proc_sig_send_unlink(Process *c_p, ErtsLink *lnk);
  *
  * This function is used instead of erts_proc_sig_send_link_exit()
  * when the signal arrives via the distribution and
- * no link structure is available.
+ * therefore no link structure is available.
  *
  * @param[in]     dep           Distribution entry of channel
  *                              that the signal arrived on.
@@ -291,6 +413,11 @@ erts_proc_sig_send_unlink(Process *c_p, ErtsLink *lnk);
  *
  * @param[in]     to            Identifier of receiver.
  *
+ * @param[in]     dist_ext      The exit reason in external term format
+ *
+ * @param[in]     hfrag         Heap frag with trace token and dist_ext
+ *                              iff available, otherwise NULL.
+ *
  * @param[in]     reason        Exit reason.
  *
  * @param[in]     token         Seq trace token.
@@ -299,15 +426,16 @@ erts_proc_sig_send_unlink(Process *c_p, ErtsLink *lnk);
 void
 erts_proc_sig_send_dist_link_exit(struct dist_entry_ *dep,
                                   Eterm from, Eterm to,
+                                  ErtsDistExternal *dist_ext,
+                                  ErlHeapFragment *hfrag,
                                   Eterm reason, Eterm token);
 
 /**
  *
- * @brief Send an unlink signal to a process.
+ * @brief Send an unlink signal to a local process.
  *
  * This function is used instead of erts_proc_sig_send_unlink()
- * when the signal arrives via the distribution and
- * no link structure is available.
+ * when the signal arrives via the distribution.
  *
  * @param[in]     dep           Distribution entry of channel
  *                              that the signal arrived on.
@@ -316,10 +444,37 @@ erts_proc_sig_send_dist_link_exit(struct dist_entry_ *dep,
  *
  * @param[in]     to            Identifier of receiver.
  *
+ * @param[in]     id            Identifier of unlink operation.
  */
 void
-erts_proc_sig_send_dist_unlink(struct dist_entry_ *dep,
-                               Eterm from, Eterm to);
+erts_proc_sig_send_dist_unlink(DistEntry *dep, Uint32 conn_id,
+                               Eterm from, Eterm to, Uint64 id);
+
+/**
+ *
+ * @brief Send an unlink acknowledgment signal to a local process.
+ *
+ * This function is used instead of erts_proc_sig_send_unlink_ack()
+ * when the signal arrives via the distribution.
+ *
+ * @param[in]     c_p           Pointer to process struct of
+ *                              currently executing process or
+ *                              NULL if not called in the context
+ *                              of an executing process.
+ *
+ * @param[in]     dep           Distribution entry of channel
+ *                              that the signal arrived on.
+ *
+ * @param[in]     from          Identifier of sender.
+ *
+ * @param[in]     to            Identifier of receiver.
+ *
+ * @param[in]     id            Identifier of unlink operation.
+ */
+void
+erts_proc_sig_send_dist_unlink_ack(Process *c_p, DistEntry *dep,
+                                   Uint32 conn_id, Eterm from, Eterm to,
+                                   Uint64 id);
 
 /**
  *
@@ -380,7 +535,7 @@ erts_proc_sig_send_monitor(ErtsMonitor *mon, Eterm to);
  *
  * This function is used instead of erts_proc_sig_send_monitor_down()
  * when the signal arrives via the distribution and
- * no link structure is available.
+ * therefore no monitor structure is available.
  *
  * @param[in]     dep           Pointer to distribution entry
  *                              of channel that the signal
@@ -392,12 +547,19 @@ erts_proc_sig_send_monitor(ErtsMonitor *mon, Eterm to);
  *
  * @param[in]     to            Identifier of receiver.
  *
+ * @param[in]     dist_ext      The exit reason in external term format
+ *
+ * @param[in]     hfrag         Heap frag with trace token and dist_ext
+ *                              iff available, otherwise NULL.
+ *
  * @param[in]     reason        Exit reason.
  *
  */
 void
 erts_proc_sig_send_dist_monitor_down(DistEntry *dep, Eterm ref,
                                      Eterm from, Eterm to,
+                                     ErtsDistExternal *dist_ext,
+                                     ErlHeapFragment *hfrag,
                                      Eterm reason);
 
 /**
@@ -506,8 +668,12 @@ erts_proc_sig_send_group_leader(Process *c_p, Eterm to, Eterm gl,
  *                              message to the sending
  *                              process (i.e., c_p).
  *
+ * @returns                     A value != 0 if the request
+ *                              was sent; otherwise, 0. If
+ *                              the request was not sent the
+ *                              process was non-existing.
  */
-void
+int
 erts_proc_sig_send_is_alive_request(Process *c_p, Eterm to,
                                     Eterm ref);
 
@@ -623,6 +789,9 @@ erts_proc_sig_send_sync_suspend(Process *c_p, Eterm to,
  * exist. The signal was not sent, and no specific
  * receive has to be entered by the caller.
  *
+ * Minimum priority, that the signal will execute under,
+ * will equal the priority of the calling process (c_p).
+ *
  * @param[in]     c_p           Pointer to process struct of
  *                              currently executing process.
  *
@@ -658,6 +827,87 @@ erts_proc_sig_send_rpc_request(Process *c_p,
                                int reply,
                                Eterm (*func)(Process *, void *, int *, ErlHeapFragment **),
                                void *arg);
+/**
+ *
+ * @brief Send an 'rpc' signal to a process.
+ *
+ * The function 'func' will be executed in the
+ * context of the receiving process. A response
+ * message '{Ref, Result}' is sent to the sender
+ * when 'func' has been called. 'Ref' is the reference
+ * returned by this function and 'Result' is the
+ * term returned by 'func'. If the return value of
+ * 'func' is not an immediate term, 'func' has to
+ * allocate a heap fragment where the result is stored
+ * and update the the heap fragment pointer pointer
+ * passed as third argument to point to it.
+ *
+ * If this function returns a reference, 'func' will
+ * be called in the context of the receiver. However,
+ * note that this might happen when the receiver is in
+ * an exiting state. The caller of this function
+ * *unconditionally* has to enter a receive that match
+ * on the returned reference in all clauses as next
+ * receive; otherwise, bad things will happen!
+ *
+ * If THE_NON_VALUE is returned, the receiver did not
+ * exist. The signal was not sent, and no specific
+ * receive has to be entered by the caller.
+ *
+ * @param[in]     c_p           Pointer to process struct of
+ *                              currently executing process.
+ *
+ * @param[in]     to            Identifier of receiver process.
+ *
+ * @param[in]     reply         Non-zero if a reply is wanted.
+ *
+ * @param[in]     func          Function to execute in the
+ *                              context of the receiver.
+ *                              First argument will be a
+ *                              pointer to the process struct
+ *                              of the receiver process.
+ *                              Second argument will be 'arg'
+ *                              (see below). Third argument
+ *                              will be a pointer to a pointer
+ *                              to a heap fragment for storage
+ *                              of result returned from 'func'
+ *                              (i.e. an 'out' parameter).
+ *
+ * @param[in]     arg           Void pointer to argument
+ *                              to pass as second argument
+ *                              in call of 'func'.
+ *
+ * @param[in]     prio          Minimum priority that the
+ *                              signal will execute under.
+ *                              Either PRIORITY_MAX,
+ *                              PRIORITY_HIGH, PRIORITY_NORMAL,
+ *                              PRIORITY_LOW, or a negative
+ *                              value. A negative value will
+ *                              cause a minimum priority that
+ *                              equals the priority of the
+ *                              calling process (c_p).
+ *
+ * @returns                     If the request was sent,
+ *                              an internal ordinary
+ *                              reference; otherwise,
+ *                              THE_NON_VALUE (non-existing
+ *                              receiver).
+ */
+Eterm
+erts_proc_sig_send_rpc_request_prio(Process *c_p,
+                                    Eterm to,
+                                    int reply,
+                                    Eterm (*func)(Process *, void *, int *, ErlHeapFragment **),
+                                    void *arg,
+                                    int prio);
+
+int
+erts_proc_sig_send_dist_spawn_reply(Eterm node,
+                                    Eterm ref,
+                                    Eterm to,
+                                    ErtsLink *lnk,
+                                    Eterm result,
+                                    Eterm token);
 
 /*
  * End of send operations of currently supported process signals.
@@ -732,6 +982,8 @@ erts_proc_sig_handle_incoming(Process *c_p, erts_aint32_t *statep,
  *                              used by the call. On output, the
  *                              amount of reductions consumed.
  *
+ * @param[in,out] pe_ctxtp      Process exit context pointer.
+ *
  * @return                      Returns a non-zero value, when
  *                              no more signals to handle in the
  *                              middle queue remain. A zero
@@ -740,7 +992,8 @@ erts_proc_sig_handle_incoming(Process *c_p, erts_aint32_t *statep,
  *                              queue.
  */
 int
-erts_proc_sig_handle_exit(Process *c_p, int *redsp);
+erts_proc_sig_handle_exit(Process *c_p, Sint *redsp,
+                          ErtsProcExitContext *pe_ctxt_p);
 
 /**
  *
@@ -961,6 +1214,54 @@ erts_proc_sig_clear_seq_trace_tokens(Process *c_p);
 void
 erts_proc_sig_handle_pending_suspend(Process *c_p);
 
+/**
+ *
+ * @brief Decode the reason term in an external signal
+ *
+ * Any distributed signal with a payload only has the control
+ * message decoded by the dist entry. The final decode of the
+ * payload is done by the process when it inspects the signal
+ * by calling this function.
+ *
+ * This functions handles both messages and link/monitor exits.
+ *
+ * Return true if the decode was successful, false otherwise.
+ *
+ * @param[in]   c_p             Pointer to executing process
+ *
+ * @param[in]   proc_lock       Locks held by process. Should always be MAIN.
+ *
+ * @param[in]   msgp            The signal to decode
+ *
+ * @param[in]   force_off_heap  If the term should be forced to be off-heap
+ */
+int
+erts_proc_sig_decode_dist(Process *proc, ErtsProcLocks proc_locks,
+                          ErtsMessage *msgp, int force_off_heap);
+
+/**
+ *
+ * @brief Check if a newly scheduled process needs to wait for
+ *        ongoing dirty handling of signals to complete.
+ *
+ * @param[in]   c_p             Pointer to executing process.
+ *
+ * @param[in]   state_in        State of process.
+ *
+ * @return                      Updated state of process if
+ *                              we had to wait; otherwise,
+ *                              state_in.
+ */
+ERTS_GLB_INLINE erts_aint32_t
+erts_proc_sig_check_wait_dirty_handle_signals(Process *c_p,
+                                              erts_aint32_t state_in);
+
+void erts_proc_sig_do_wait_dirty_handle_signals__(Process *c_p);
+
+
+ErtsDistExternal *
+erts_proc_sig_get_external(ErtsMessage *msgp);
+
 /**
  * @brief Initialize this functionality
  */
@@ -970,8 +1271,9 @@ void
 erts_proc_sig_debug_foreach_sig(Process *c_p,
                                 void (*msg_func)(ErtsMessage *, void *),
                                 void (*oh_func)(ErlOffHeap *, void *),
-                                void (*mon_func)(ErtsMonitor *, void *),
-                                void (*lnk_func)(ErtsLink *, void *),
+                                ErtsMonitorFunc mon_func,
+                                ErtsLinkFunc lnk_func,
+                                void (*ext_func)(ErtsDistExternal *, void *),
                                 void *arg);
 
 extern Process *erts_dirty_process_signal_handler;
@@ -983,6 +1285,18 @@ Sint erts_proc_sig_fetch_msgq_len_offs__(Process *proc);
 
 #if ERTS_GLB_INLINE_INCL_FUNC_DEF
 
+ERTS_GLB_INLINE Uint64
+erts_proc_sig_new_unlink_id(Process *c_p)
+{
+    Uint64 id;
+    ASSERT(c_p);
+
+    id = (Uint64) c_p->uniq++;
+    if (id == 0)
+        id = (Uint64) c_p->uniq++;
+    return id;
+}
+
 ERTS_GLB_INLINE Sint
 erts_proc_sig_fetch(Process *proc)
 {
@@ -996,6 +1310,8 @@ erts_proc_sig_fetch(Process *proc)
                        == (ERTS_PROC_LOCK_MAIN
                            | ERTS_PROC_LOCK_MSGQ)));
 
+    ASSERT(!(proc->sig_qs.flags & FS_HANDLING_SIGS));
+
     ERTS_HDBG_CHECK_SIGNAL_IN_QUEUE(proc);
     ERTS_HDBG_CHECK_SIGNAL_PRIV_QUEUE(proc, !0);
 
@@ -1039,12 +1355,33 @@ erts_proc_notify_new_sig(Process* rp, erts_aint32_t state,
         state = erts_proc_sys_schedule(rp, state, enable_flag);
     }
 
-    if (state & (ERTS_PSFLG_DIRTY_RUNNING
-                 | ERTS_PSFLG_DIRTY_RUNNING_SYS)) {
+    if (state & ERTS_PSFLG_DIRTY_RUNNING) {
+        /*
+         * We ignore ERTS_PSFLG_DIRTY_RUNNING_SYS. For
+         * more info see erts_execute_dirty_system_task()
+         * in erl_process.c.
+         */
         erts_make_dirty_proc_handled(rp->common.id, state, -1);
     }
 }
 
+ERTS_GLB_INLINE erts_aint32_t
+erts_proc_sig_check_wait_dirty_handle_signals(Process *c_p,
+                                              erts_aint32_t state_in)
+{
+    erts_aint32_t state = state_in;
+    ASSERT(erts_get_scheduler_data()->type == ERTS_SCHED_NORMAL);
+    ERTS_LC_ASSERT(ERTS_PROC_LOCK_MAIN == erts_proc_lc_my_proc_locks(c_p));
+
+    if (c_p->sig_qs.flags & FS_HANDLING_SIGS) {
+        erts_proc_sig_do_wait_dirty_handle_signals__(c_p);
+        state = erts_atomic32_read_mb(&c_p->state);
+    }
+    ERTS_LC_ASSERT(ERTS_PROC_LOCK_MAIN == erts_proc_lc_my_proc_locks(c_p));
+    ASSERT(!(c_p->sig_qs.flags & FS_HANDLING_SIGS));
+    return state;
+}
+
 #endif /* ERTS_GLB_INLINE_INCL_FUNC_DEF */
 
 #endif /* ERTS_PROC_SIG_QUEUE_H__ */
