@@ -1,7 +1,7 @@
 /*
  * %CopyrightBegin%
  *
- * Copyright Ericsson AB 1999-2018. All Rights Reserved.
+ * Copyright Ericsson AB 1999-2020. All Rights Reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -635,9 +635,11 @@ write_sys_msg_to_port(Eterm unused_to,
 		      Eterm message) {
     byte *buffer;
     byte *ptr;
-    unsigned size;
+    Uint size;
+
+    if (erts_encode_ext_size(message, &size) != ERTS_EXT_SZ_OK)
+	erts_exit(ERTS_ERROR_EXIT, "Internal error: System limit\n");
 
-    size = erts_encode_ext_size(message);
     buffer = (byte *) erts_alloc(ERTS_ALC_T_TMP, size);
 
     ptr = buffer;
@@ -711,7 +713,9 @@ trace_sched(Process *p, ErtsProcLocks locks, Eterm what)
     trace_sched_aux(p, locks, what);
 }
 
-/* Send {trace_ts, Pid, Send, Msg, DestPid, Timestamp}
+/* Send {trace_ts, Pid, Send, Msg, DestPid, PamResult, Timestamp}
+ * or   {trace_ts, Pid, Send, Msg, DestPid, Timestamp}
+ * or   {trace, Pid, Send, Msg, DestPid, PamResult}
  * or   {trace, Pid, Send, Msg, DestPid}
  *
  * where 'Send' is 'send' or 'send_to_non_existing_process'.
@@ -771,7 +775,9 @@ trace_send(Process *p, Eterm to, Eterm msg)
     erts_match_set_release_result_trace(p, pam_result);
 }
 
-/* Send {trace_ts, Pid, receive, Msg, Timestamp}
+/* Send {trace_ts, Pid, receive, Msg, PamResult, Timestamp}
+ * or   {trace_ts, Pid, receive, Msg, Timestamp}
+ * or   {trace, Pid, receive, Msg, PamResult}
  * or   {trace, Pid, receive, Msg}
  */
 void
@@ -824,7 +830,7 @@ trace_receive(Process* receiver,
 }
 
 int
-seq_trace_update_send(Process *p)
+seq_trace_update_serial(Process *p)
 {
     ErtsTracer seq_tracer = erts_get_system_seq_tracer();
     ASSERT((is_tuple(SEQ_TRACE_TOKEN(p)) || is_nil(SEQ_TRACE_TOKEN(p))));
@@ -847,6 +853,18 @@ seq_trace_update_send(Process *p)
     return 1;
 }
 
+void
+erts_seq_trace_update_node_token(Eterm token)
+{
+    Eterm serial;
+    Uint serial_num;
+    SEQ_TRACE_T_SENDER(token) = erts_this_dist_entry->sysname;
+    serial = SEQ_TRACE_T_SERIAL(token);
+    serial_num = unsigned_val(serial);
+    serial_num++;
+    SEQ_TRACE_T_SERIAL(token) = make_small(serial_num);
+}
+
 
 /* Send a sequential trace message to the sequential tracer.
  * p is the caller (which contains the trace token), 
@@ -923,6 +941,9 @@ seq_trace_output_generic(Eterm token, Eterm msg, Uint type,
 #undef LOCAL_HEAP_SIZE
 }
 
+
+
+
 /* Send {trace_ts, Pid, return_to, {Mod, Func, Arity}, Timestamp}
  * or   {trace, Pid, return_to, {Mod, Func, Arity}}
  */
@@ -1342,6 +1363,7 @@ trace_gc(Process *p, Eterm what, Uint size, Eterm msg)
     Eterm* hp;
     Uint sz = 0;
     Eterm tup;
+    ErtsThrPrgrDelayHandle dhndl = erts_thr_progress_unmanaged_delay();
 
     if (is_tracer_enabled(p, ERTS_PROC_LOCK_MAIN, &p->common, &tnif,
                           TRACE_FUN_E_GC, what)) {
@@ -1361,6 +1383,7 @@ trace_gc(Process *p, Eterm what, Uint size, Eterm msg)
         if (o_hp)
             erts_free(ERTS_ALC_T_TMP, o_hp);
     }
+    erts_thr_progress_unmanaged_continue(dhndl);
 }
 
 void 
@@ -2112,36 +2135,49 @@ sys_msg_disp_failure(ErtsSysMsgQ *smqp, Eterm receiver)
 	erts_thr_progress_unblock();
 	break;
     case SYS_MSG_TYPE_ERRLGR: {
-	char *no_elgger = "(no logger present)";
 	Eterm *tp;
 	Eterm tag;
+
 	if (is_not_tuple(smqp->msg)) {
-	unexpected_elmsg:
-	    erts_fprintf(stderr,
-			 "%s unexpected logger message: %T\n",
-			 no_elgger,
-			 smqp->msg);
+	    goto unexpected_error_msg;
+	}
+	tp = tuple_val(smqp->msg);
+	if (arityval(tp[0]) != 2) {
+	    goto unexpected_error_msg;
+	}
+	if (is_not_tuple(tp[2])) {
+	    goto unexpected_error_msg;
+	}
+	tp = tuple_val(tp[2]);
+	if (arityval(tp[0]) != 3) {
+	    goto unexpected_error_msg;
+	}
+	tag = tp[1];
+	if (is_not_tuple(tp[3])) {
+	    goto unexpected_error_msg;
+	}
+	tp = tuple_val(tp[3]);
+	if (arityval(tp[0]) != 3) {
+	    goto unexpected_error_msg;
+	}
+	if (is_not_list(tp[3])) {
+	    goto unexpected_error_msg;
 	}
 
-	tp = tuple_val(smqp->msg);
-	if (arityval(tp[0]) != 2)
-	    goto unexpected_elmsg;
-	if (is_not_tuple(tp[2]))
-	    goto unexpected_elmsg;
-	tp = tuple_val(tp[2]);
-	if (arityval(tp[0]) != 3)
-	    goto unexpected_elmsg;
-	tag = tp[1];
-	if (is_not_tuple(tp[3]))
-	    goto unexpected_elmsg;
-	tp = tuple_val(tp[3]);
-	if (arityval(tp[0]) != 3)
-	    goto unexpected_elmsg;
-	if (is_not_list(tp[3]))
-	    goto unexpected_elmsg;
-	erts_fprintf(stderr, "%s %T: %T\n",
-		     no_elgger, tag, CAR(list_val(tp[3])));
-	break;
+        {
+            static const char *no_logger = "(no logger present)";
+        /* no_error_logger: */
+            erts_fprintf(stderr, "%s %T: %T\n",
+                         no_logger, tag, CAR(list_val(tp[3])));
+            break;
+        unexpected_error_msg:
+            erts_fprintf(stderr,
+                         "%s unexpected logger message: %T\n",
+                         no_logger,
+                         smqp->msg);
+            break;
+        }
+        ASSERT(0);
     }
     case SYS_MSG_TYPE_PROC_MSG:
         break;
@@ -2188,11 +2224,12 @@ sys_msg_dispatcher_wait(void *vwait_p)
     erts_mtx_unlock(&smq_mtx);
 }
 
+static ErtsSysMsgQ *local_sys_message_queue = NULL;
+
 static void *
 sys_msg_dispatcher_func(void *unused)
 {
     ErtsThrPrgrCallbacks callbacks;
-    ErtsSysMsgQ *local_sys_message_queue = NULL;
     ErtsThrPrgrData *tpd;
     int wait = 0;
 
@@ -2200,13 +2237,15 @@ sys_msg_dispatcher_func(void *unused)
     erts_lc_set_thread_name("system message dispatcher");
 #endif
 
+    local_sys_message_queue = NULL;
+
     callbacks.arg = (void *) &wait;
     callbacks.wakeup = sys_msg_dispatcher_wakeup;
     callbacks.prepare_wait = sys_msg_dispatcher_prep_wait;
     callbacks.wait = sys_msg_dispatcher_wait;
     callbacks.finalize_wait = sys_msg_dispatcher_fin_wait;
 
-    tpd = erts_thr_progress_register_managed_thread(NULL, &callbacks, 0);
+    tpd = erts_thr_progress_register_managed_thread(NULL, &callbacks, 0, 0);
 
     while (1) {
 	int end_wait = 0;
@@ -2225,6 +2264,7 @@ sys_msg_dispatcher_func(void *unused)
 
 	/* Fetch current trace message queue ... */
 	if (!sys_message_queue) {
+            wait = 1;
 	    erts_mtx_unlock(&smq_mtx);
 	    end_wait = 1;
 	    erts_thr_progress_active(tpd, 0);
@@ -2232,8 +2272,23 @@ sys_msg_dispatcher_func(void *unused)
 	    erts_mtx_lock(&smq_mtx);
 	}
 
-	while (!sys_message_queue)
-	    erts_cnd_wait(&smq_cnd, &smq_mtx);
+	while (!sys_message_queue) {
+            if (wait)
+                erts_cnd_wait(&smq_cnd, &smq_mtx);
+            if (sys_message_queue)
+                break;
+            wait = 1;
+	    erts_mtx_unlock(&smq_mtx);
+            /*
+             * Ensure thread progress continue. We might have
+             * been the last thread to go to sleep. In that case
+             * erts_thr_progress_finalize_wait() will take care
+             * of it...
+             */
+	    erts_thr_progress_finalize_wait(tpd);
+	    erts_thr_progress_prepare_wait(tpd);
+	    erts_mtx_lock(&smq_mtx);
+        }
 
 	local_sys_message_queue = sys_message_queue;
 	sys_message_queue = NULL;
@@ -2256,6 +2311,8 @@ sys_msg_dispatcher_func(void *unused)
 	    Process *proc = NULL;
 	    Port *port = NULL;
 
+            ASSERT(is_value(smqp->msg));
+
 	    if (erts_thr_progress_update(tpd))
 		erts_thr_progress_leader_update(tpd);
 
@@ -2368,6 +2425,7 @@ sys_msg_dispatcher_func(void *unused)
 		erts_fprintf(stderr, "dropped\n");
 #endif
 	    }
+            smqp->msg = THE_NON_VALUE;
 	}
     }
 
@@ -2375,32 +2433,38 @@ sys_msg_dispatcher_func(void *unused)
 }
 
 void
-erts_foreach_sys_msg_in_q(void (*func)(Eterm,
-				       Eterm,
-				       Eterm,
-				       ErlHeapFragment *))
+erts_debug_foreach_sys_msg_in_q(void (*func)(Eterm,
+                                             Eterm,
+                                             Eterm,
+                                             ErlHeapFragment *))
 {
-    ErtsSysMsgQ *sm;
-    erts_mtx_lock(&smq_mtx);
-    for (sm = sys_message_queue; sm; sm = sm->next) {
-	Eterm to;
-	switch (sm->type) {
-	case SYS_MSG_TYPE_SYSMON:
-	    to = erts_get_system_monitor();
-	    break;
-	case SYS_MSG_TYPE_SYSPROF:
-	    to = erts_get_system_profile();
-	    break;
-	case SYS_MSG_TYPE_ERRLGR:
-	    to = erts_get_system_logger();
-	    break;
-	default:
-	    to = NIL;
-	    break;
-	}
-	(*func)(sm->from, to, sm->msg, sm->bp);
+    ErtsSysMsgQ *smq[] = {sys_message_queue, local_sys_message_queue};
+    int i;
+
+    ERTS_LC_ASSERT(erts_thr_progress_is_blocking());
+
+    for (i = 0; i < sizeof(smq)/sizeof(smq[0]); i++) {
+        ErtsSysMsgQ *sm;
+        for (sm = smq[i]; sm; sm = sm->next) {
+            Eterm to;
+            switch (sm->type) {
+            case SYS_MSG_TYPE_SYSMON:
+                to = erts_get_system_monitor();
+                break;
+            case SYS_MSG_TYPE_SYSPROF:
+                to = erts_get_system_profile();
+                break;
+            case SYS_MSG_TYPE_ERRLGR:
+                to = erts_get_system_logger();
+                break;
+            default:
+                to = NIL;
+                break;
+            }
+            if (is_value(sm->msg))
+                (*func)(sm->from, to, sm->msg, sm->bp);
+        }
     }
-    erts_mtx_unlock(&smq_mtx);
 }
 
 
@@ -2570,6 +2634,7 @@ lookup_tracer_nif(const ErtsTracer tracer)
     ErtsTracerNif tnif_tmpl;
     ErtsTracerNif *tnif;
     tnif_tmpl.module = ERTS_TRACER_MODULE(tracer);
+    ERTS_LC_ASSERT(erts_thr_progress_lc_is_delaying() || erts_get_scheduler_id() > 0);
     erts_rwmtx_rlock(&tracer_mtx);
     if ((tnif = hash_get(tracer_hash, &tnif_tmpl)) == NULL) {
         erts_rwmtx_runlock(&tracer_mtx);
