@@ -1,7 +1,7 @@
 %%--------------------------------------------------------------------
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2013-2018. All Rights Reserved.
+%% Copyright Ericsson AB 2013-2021. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -31,6 +31,7 @@
 -include_lib("common_test/src/ct_netconfc.hrl").
 -include("netconfc_test_lib.hrl").
 
+-compile(nowarn_export_all).
 -compile(export_all).
 
 suite() ->
@@ -98,6 +99,7 @@ all() ->
 	     get_event_streams,
 	     create_subscription,
 	     receive_one_event,
+             receive_two_events,
 	     receive_multiple_events,
 	     receive_event_and_rpc,
 	     receive_event_and_rpc_in_chunks,
@@ -205,7 +207,7 @@ hello_required_exists(Config) ->
     SshDir = ?config(ssh_dir,Config),
     {ok,_Client1} = open_configured_success(my_named_connection,SshDir),
 
-    %% Check that same name can not be used twice
+    %% Check that same name cannot be used twice
     {error,{connection_exists,_Client1}} =
 	ct_netconfc:open(my_named_connection,[{user_dir,SshDir}]),
 
@@ -271,7 +273,7 @@ no_client_hello(Config) ->
 
     %% Tell server to receive a get request and then die without
     %% replying since no hello has been received. (is this correct
-    %% behavoiur??)
+    %% behaviour??)
     ?NS:expect_do(get,close),
     {error,closed} = ct_netconfc:get(Client,whatever),
     ok.
@@ -385,7 +387,7 @@ timeout_get(Config) ->
 %% received, the timeout message might already be sent when the timer
 %% is cancelled. This test checks that the timeout message is flushed
 %% from the message queue. If it isn't, the client crashes and the
-%% session can not be closed afterwards.
+%% session cannot be closed afterwards.
 %% Note that we can only hope that the test case triggers the problem
 %% every now and then, as it is very timing dependent...
 flush_timeout_get(Config) ->
@@ -979,32 +981,44 @@ create_subscription(Config) ->
     ok.
 
 receive_one_event(Config) ->
+    receive_events(Config, 1, []).
+
+receive_two_events(Config) ->
+    receive_events(Config, 2, [{receiver, self()}, {receiver, self()}]).
+
+receive_events(Config, N, Opts) ->
     SshDir = ?config(ssh_dir,Config),
-    {ok,Client} = open_success(SshDir),
+    {ok,Client} = open_success(SshDir, Opts),
     ?NS:expect_reply({'create-subscription',[stream]},ok),
     ?ok = ct_netconfc:create_subscription(Client),
 
     ?NS:hupp({send_events,1}),
 
-    receive
-	%% Matching ?NS:make_msg({event,_})
-	{notification,?NETCONF_NOTIF_NAMESPACE_ATTR,
-	 [{eventTime,[],[_Time]},
-	  {event,[{xmlns,"http://my.namespaces.com/event"}],
-	   [{severity,_,_},
-	    {description,_,_}]}]} ->
-	    ok;
-	Other ->
-	    ct:fail({got_unexpected_while_waiting_for_event, Other})
-    after 3000 ->
-	    ct:fail(timeout_waiting_for_event)
-    end,
+    receive_events(N),
 
     ?NS:expect_do_reply('close-session',close,ok),
     ?ok = ct_netconfc:close_session(Client),
 
     ok.
 
+receive_events(0) ->
+    ok;
+receive_events(N) ->
+    receive
+        {notification,?NETCONF_NOTIF_NAMESPACE_ATTR,
+         [{eventTime,[],[_Time]},
+          {event,[{xmlns,"http://my.namespaces.com/event"}],
+           [{severity,_,_},
+            {description,_,_}]}]} ->
+            ok;
+        Other ->
+            ct:fail({got_unexpected_while_waiting_for_event, Other})
+    after 3000 ->
+            ct:fail(timeout_waiting_for_event)
+    end,
+    receive_events(N-1).
+
+
 receive_multiple_events(Config) ->
     SshDir = ?config(ssh_dir,Config),
     {ok,Client} = open_success(SshDir),
