@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2017-2018. All Rights Reserved.
+%% Copyright Ericsson AB 2017-2022. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -23,8 +23,61 @@
 
 -include_lib("common_test/include/ct.hrl").
 
-%% Note: This directive should only be used in test suites.
--compile(export_all).
+-export([
+         suite/0,
+         all/0,
+         groups/0,
+         init_per_suite/1,
+         end_per_suite/1,
+         init_per_group/2,
+         end_per_group/2,
+         init_per_testcase/2,
+         end_per_testcase/2,
+         get_all_possible_methods/0,
+         get_all_possible_methods/1,
+         engine_load_register_method/0,
+         engine_load_register_method/1,
+         engine_list/0,
+         engine_list/1,
+         get_id_and_name/0,
+         get_id_and_name/1,
+         engine_by_id/0,
+         engine_by_id/1,
+         bad_arguments/0,
+         bad_arguments/1,
+         unknown_engine/0,
+         unknown_engine/1,
+         pre_command_fail_bad_value/0,
+         pre_command_fail_bad_value/1,
+         pre_command_fail_bad_key/0,
+         pre_command_fail_bad_key/1,
+         failed_engine_init/0,
+         failed_engine_init/1,
+         ctrl_cmd_string/0,
+         ctrl_cmd_string/1,
+         ctrl_cmd_string_optional/0,
+         ctrl_cmd_string_optional/1,
+         ensure_load/0,
+         ensure_load/1,
+	 gc_clean/0,
+         gc_clean/1,
+         sign_verify_rsa/1,
+         sign_verify_rsa_fake/1,
+         sign_verify_dsa/1,
+         sign_verify_ecdsa/1,
+         sign_verify_rsa_pwd/1,
+         sign_verify_rsa_pwd_bad_pwd/1,
+         priv_encrypt_pub_decrypt_rsa/1,
+         priv_encrypt_pub_decrypt_rsa_pwd/1,
+         pub_encrypt_priv_decrypt_rsa/1,
+         pub_encrypt_priv_decrypt_rsa_pwd/1,
+         get_pub_from_priv_key_rsa/1,
+         get_pub_from_priv_key_rsa_pwd/1,
+         get_pub_from_priv_key_rsa_pwd_no_pwd/1,
+         get_pub_from_priv_key_rsa_pwd_bad_pwd/1,
+         get_pub_from_priv_key_dsa/1,
+         get_pub_from_priv_key_ecdsa/1
+        ]).
 
 %%--------------------------------------------------------------------
 %% Common Test interface functions -----------------------------------
@@ -37,9 +90,7 @@ suite() ->
 all() ->
     [
      get_all_possible_methods,
-     engine_load_all_methods,
-     engine_load_some_methods,
-     multiple_engine_load,
+     engine_load_register_method,
      engine_list,
      get_id_and_name,
      engine_by_id,
@@ -51,6 +102,7 @@ all() ->
      ctrl_cmd_string,
      ctrl_cmd_string_optional,
      ensure_load,
+     gc_clean,
      {group, engine_stored_key},
      {group, engine_fakes_rsa}
     ].
@@ -80,7 +132,7 @@ groups() ->
 
 
 init_per_suite(Config) ->
-    case {os:type(), crypto:info_lib()} of
+    try {os:type(), crypto:info_lib()} of
         {_, [{_,_, <<"OpenSSL 1.0.1s-freebsd  1 Mar 2016">>}]} ->
             {skip, "Problem with engine on OpenSSL 1.0.1s-freebsd"};
 
@@ -100,6 +152,8 @@ init_per_suite(Config) ->
             catch _:_ ->
                     {skip, "Crypto did not start"}
             end
+    catch _:_ ->
+	    {skip, "Crypto not loaded"}
     end.
 
 end_per_suite(_Config) ->
@@ -118,10 +172,10 @@ init_per_group(engine_fakes_rsa, Config) ->
 init_per_group(_Group, Config0) ->
     Config0.
 
-
 group_load_engine(Config, ExcludeMthds) ->
-    case load_storage_engine(Config, ExcludeMthds) of
+    case load_storage_engine(Config) of
         {ok, E} ->
+	    ok = crypto:engine_register(E, crypto:engine_get_all_methods() -- ExcludeMthds),
             KeyDir = key_dir(Config),
             [{storage_engine,E}, {storage_dir,KeyDir} | Config];
         {error, notexist} ->
@@ -135,10 +189,6 @@ group_load_engine(Config, ExcludeMthds) ->
             {fail, "Engine load failed"}
     end.
 
-
-
-
-
 end_per_group(_, Config) ->
     case proplists:get_value(storage_engine, Config) of
         undefined ->
@@ -148,8 +198,21 @@ end_per_group(_, Config) ->
     end.
 
 %%--------------------------------------------------------------------
-init_per_testcase(_Case, Config) ->
-    Config.
+init_per_testcase(Case, Config) ->
+    case string:tokens(atom_to_list(Case),"_") of
+        ["sign","verify",Type|_] ->
+            skip_if_unsup(list_to_atom(Type), Config);
+
+        ["priv","encrypt","pub","decrypt",Type|_] ->
+            skip_if_unsup(list_to_atom(Type), Config);
+
+        ["get","pub","from","priv","key",Type|_] ->
+            skip_if_unsup(list_to_atom(Type), Config);
+
+        _ ->
+            Config
+    end.
+
 end_per_testcase(_Case, _Config) ->
     ok.
 
@@ -170,12 +233,13 @@ get_all_possible_methods(Config) when is_list(Config) ->
             {skip, "Engine not supported on this SSL version"}
     end.
 
-engine_load_all_methods()->
+%%-------------------------------------------------------------------------
+engine_load_register_method()->
     [{doc, "Use a dummy md5 engine that does not implement md5"
       "but rather returns a static binary to test that crypto:engine_load "
       "functions works."}].
 
-engine_load_all_methods(Config) when is_list(Config) ->
+engine_load_register_method(Config) when is_list(Config) ->
     case crypto:get_test_engine() of
         {error, notexist} ->
             {skip, "OTP Test engine not found"};
@@ -187,8 +251,10 @@ engine_load_all_methods(Config) when is_list(Config) ->
                 case crypto:engine_load(<<"dynamic">>,
                                         [{<<"SO_PATH">>, Engine},
                                          <<"LOAD">>],
-                                        []) of
+                                        []) of		    
                     {ok, E} ->
+			ok = crypto:engine_register(E, [engine_method_digests]),
+
                         case crypto:hash(md5, "Don't panic") of
                             Md5Hash1 ->
                                 ct:fail(fail_to_load_still_original_engine);
@@ -197,7 +263,10 @@ engine_load_all_methods(Config) when is_list(Config) ->
                             _ ->
                                 ct:fail(fail_to_load_engine)
                         end,
+
+			ok = crypto:engine_unregister(E, [engine_method_digests]),
                         ok = crypto:engine_unload(E),
+
                         case crypto:hash(md5, "Don't panic") of
                             Md5Hash2 ->
                                 ct:fail(fail_to_unload_still_test_engine);
@@ -215,127 +284,7 @@ engine_load_all_methods(Config) when is_list(Config) ->
            end
     end.
 
-engine_load_some_methods()->
-    [{doc, "Use a dummy md5 engine that does not implement md5"
-      "but rather returns a static binary to test that crypto:engine_load "
-      "functions works."}].
-
-engine_load_some_methods(Config) when is_list(Config) ->
-    case crypto:get_test_engine() of
-        {error, notexist} ->
-            {skip, "OTP Test engine not found"};
-        {ok, Engine} ->
-            try
-                Md5Hash1 =  <<106,30,3,246,166,222,229,158,244,217,241,179,50,232,107,109>>,
-                Md5Hash1 = crypto:hash(md5, "Don't panic"),
-                Md5Hash2 =  <<0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15>>,
-                EngineMethods = crypto:engine_get_all_methods() --
-                    [engine_method_dh, engine_method_rand,
-                     engine_method_ciphers, engine_method_store,
-                     engine_method_pkey_meths, engine_method_pkey_asn1_meths],
-                case crypto:engine_load(<<"dynamic">>,
-                                        [{<<"SO_PATH">>, Engine},
-                                         <<"LOAD">>],
-                                        [],
-                                        EngineMethods) of
-                    {ok, E} ->
-                        case crypto:hash(md5, "Don't panic") of
-                            Md5Hash1 ->
-                                ct:fail(fail_to_load_engine_still_original);
-                            Md5Hash2 ->
-                                ok;
-                            _ ->
-                                ct:fail(fail_to_load_engine)
-                        end,
-                        ok = crypto:engine_unload(E),
-                        case crypto:hash(md5, "Don't panic") of
-                            Md5Hash2 ->
-                                ct:fail(fail_to_unload_still_test_engine);
-                            Md5Hash1 ->
-                                ok;
-                            _ ->
-                                ct:fail(fail_to_unload_engine)
-                        end;
-                    {error, bad_engine_id} ->
-                    {skip, "Dynamic Engine not supported"}
-                end
-           catch
-               error:notsup ->
-                  {skip, "Engine not supported on this SSL version"}
-           end
-    end.
-
-multiple_engine_load()->
-    [{doc, "Use a dummy md5 engine that does not implement md5"
-      "but rather returns a static binary to test that crypto:engine_load "
-      "functions works when called multiple times."}].
-
-multiple_engine_load(Config) when is_list(Config) ->
-    case crypto:get_test_engine() of
-        {error, notexist} ->
-            {skip, "OTP Test engine not found"};
-        {ok, Engine} ->
-            try
-                Md5Hash1 =  <<106,30,3,246,166,222,229,158,244,217,241,179,50,232,107,109>>,
-                Md5Hash1 = crypto:hash(md5, "Don't panic"),
-                Md5Hash2 =  <<0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15>>,
-                case crypto:engine_load(<<"dynamic">>,
-                                        [{<<"SO_PATH">>, Engine},
-                                         <<"LOAD">>],
-                                        []) of
-                    {ok, E} ->
-                        {ok, E1} = crypto:engine_load(<<"dynamic">>,
-                                        [{<<"SO_PATH">>, Engine},
-                                         <<"LOAD">>],
-                                        []),
-                        {ok, E2} = crypto:engine_load(<<"dynamic">>,
-                                        [{<<"SO_PATH">>, Engine},
-                                         <<"LOAD">>],
-                                        []),
-                        case crypto:hash(md5, "Don't panic") of
-                            Md5Hash1 ->
-                                ct:fail(fail_to_load_still_original_engine);
-                            Md5Hash2 ->
-                                ok;
-                            _ ->
-                                ct:fail(fail_to_load_engine)
-                        end,
-                        ok = crypto:engine_unload(E2),
-                        case crypto:hash(md5, "Don't panic") of
-                            Md5Hash1 ->
-                                ct:fail(fail_to_load_still_original_engine);
-                            Md5Hash2 ->
-                                ok;
-                            _ ->
-                                ct:fail(fail_to_load_engine)
-                        end,
-                        ok = crypto:engine_unload(E),
-                        case crypto:hash(md5, "Don't panic") of
-                            Md5Hash1 ->
-                                ct:fail(fail_to_load_still_original_engine);
-                            Md5Hash2 ->
-                                ok;
-                            _ ->
-                                ct:fail(fail_to_load_engine)
-                        end,
-                        ok = crypto:engine_unload(E1),
-                        case crypto:hash(md5, "Don't panic") of
-                            Md5Hash2 ->
-                                ct:fail(fail_to_unload_still_test_engine);
-                            Md5Hash1 ->
-                                ok;
-                            _ ->
-                                ct:fail(fail_to_unload_engine)
-                        end;
-                    {error, bad_engine_id} ->
-                        {skip, "Dynamic Engine not supported"}
-                end
-           catch
-               error:notsup ->
-                  {skip, "Engine not supported on this SSL version"}
-           end
-    end.
-
+%%-------------------------------------------------------------------------
 engine_list()->
     [{doc, "Test add and remove engine ID to the SSL internal engine list."}].
 
@@ -366,6 +315,7 @@ engine_list(Config) when is_list(Config) ->
            end
     end.
 
+%%-------------------------------------------------------------------------
 get_id_and_name()->
     [{doc, "Test fetching id and name from an engine."}].
 
@@ -392,6 +342,7 @@ get_id_and_name(Config) when is_list(Config) ->
            end
     end.
 
+%%-------------------------------------------------------------------------
 engine_by_id()->
     [{doc, "Test fetching a new reference the the engine when the"
      "engine id is added to the SSL engine list."}].
@@ -485,6 +436,7 @@ bad_arguments(Config) when is_list(Config) ->
           end
     end.
 
+%%-------------------------------------------------------------------------
 unknown_engine() ->
     [{doc, "Try to load a non existent engine."}].
 
@@ -497,6 +449,7 @@ unknown_engine(Config) when is_list(Config) ->
            {skip, "Engine not supported on this SSL version"}
     end.
 
+%%-------------------------------------------------------------------------
 pre_command_fail_bad_value() ->
     [{doc, "Test pre command due to bad value"}].
 
@@ -519,6 +472,7 @@ pre_command_fail_bad_value(Config) when is_list(Config) ->
            {skip, "Engine not supported on this SSL version"}
     end.
 
+%%-------------------------------------------------------------------------
 pre_command_fail_bad_key() ->
     [{doc, "Test pre command due to bad key"}].
 
@@ -544,6 +498,7 @@ pre_command_fail_bad_key(Config) when is_list(Config) ->
           {skip, "Engine not supported on this SSL version"}
    end.
 
+%%-------------------------------------------------------------------------
 failed_engine_init()->
     [{doc, "Test failing engine init due to missed pre command"}].
 
@@ -568,7 +523,6 @@ failed_engine_init(Config) when is_list(Config) ->
           {skip, "Engine not supported on this SSL version"}
    end.
 
-
 %%-------------------------------------------------------------------------
 %% Test the optional flag in ctrl comands
 ctrl_cmd_string()->
@@ -601,6 +555,7 @@ ctrl_cmd_string(Config) when is_list(Config) ->
           {skip, "Engine not supported on this SSL version"}
    end.
 
+%%-------------------------------------------------------------------------
 ctrl_cmd_string_optional()->
     [{doc, "Test that a not known optional ctrl comand do not fail"}].
 ctrl_cmd_string_optional(Config) when is_list(Config) ->
@@ -632,6 +587,9 @@ ctrl_cmd_string_optional(Config) when is_list(Config) ->
           {skip, "Engine not supported on this SSL version"}
    end.
 
+%%-------------------------------------------------------------------------
+%
+%
 ensure_load()->
     [{doc, "Test the special ensure load function."}].
 
@@ -641,12 +599,13 @@ ensure_load(Config) when is_list(Config) ->
             {skip, "OTP Test engine not found"};
         {ok, Engine} ->
             try
-                Md5Hash1 =  <<106,30,3,246,166,222,229,158,244,217,241,179,50,232,107,109>>,
-                Md5Hash1 = crypto:hash(md5, "Don't panic"),
-                Md5Hash2 =  <<0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15>>,
-                case crypto:ensure_engine_loaded(<<"MD5">>, Engine) of
+                Md5Hash1 = <<106,30,3,246,166,222,229,158,244,217,241,179,50,232,107,109>>,
+	        Md5Hash1 = crypto:hash(md5, "Don't panic"),
+	        Md5Hash2 =  <<0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15>>,
+	        case crypto:ensure_engine_loaded(<<"MD5">>, Engine) of
                     {ok, E} ->
-                        {ok, _E1} = crypto:ensure_engine_loaded(<<"MD5">>, Engine),
+			ok = crypto:engine_register(E, [engine_method_digests]),
+
                         case crypto:hash(md5, "Don't panic") of
                             Md5Hash1 ->
                                 ct:fail(fail_to_load_still_original_engine);
@@ -655,26 +614,159 @@ ensure_load(Config) when is_list(Config) ->
                             _ ->
                                 ct:fail(fail_to_load_engine)
                         end,
+
+                        {ok, E1} = crypto:ensure_engine_loaded(<<"MD5">>, Engine),
+                        case crypto:hash(md5, "Don't panic") of
+                            Md5Hash1 ->
+                                ct:fail(fail_to_load_still_original_engine);
+                            Md5Hash2 ->
+                                ok;
+                            _ ->
+                                ct:fail(fail_to_load_engine)
+                        end,
+
+                        ok = crypto:ensure_engine_unloaded(E1),
+                        case crypto:hash(md5, "Don't panic") of
+                            Md5Hash1 ->
+                                ct:fail(fail_test_engine_unloaded);
+                            Md5Hash2 ->
+				ok;
+                            _ ->
+                                ct:fail(fail_to_unload_engine)
+                        end,
+
+                        {ok, E2} = crypto:ensure_engine_loaded(<<"MD5">>, Engine),
+			case crypto:hash(md5, "Don't panic") of
+                            Md5Hash1 ->
+                                ct:fail(fail_test_engine_not_loaded);
+                            Md5Hash2 ->
+				ok;
+                            _ ->
+                                ct:fail(fail_to_load_engine)
+                        end,
+			ok = crypto:ensure_engine_unloaded(E2),
+
+                        {ok, E3} = crypto:ensure_engine_loaded(<<"MD5">>, Engine),
+			case crypto:hash(md5, "Don't panic") of
+                            Md5Hash1 ->
+                                ct:fail(fail_test_engine_not_loaded);
+                            Md5Hash2 ->
+				ok;
+                            _ ->
+                                ct:fail(fail_to_load_engine)
+                        end,
+
                         ok = crypto:ensure_engine_unloaded(E),
                         case crypto:hash(md5, "Don't panic") of
-                            Md5Hash2 ->
-                                ct:fail(fail_to_unload_still_test_engine);
                             Md5Hash1 ->
+                                ct:fail(fail_test_engine_not_loaded);
+                            Md5Hash2 ->
                                 ok;
                             _ ->
                                 ct:fail(fail_to_unload_engine)
+                        end,
+
+			ok = crypto:engine_unregister(E3, [engine_method_digests]),
+			ok = crypto:ensure_engine_unloaded(E3),
+                        case crypto:hash(md5, "Don't panic") of
+                            Md5Hash1 ->
+                                ok;
+                            Md5Hash2 ->
+                                ct:fail(fail_to_unload_still_test_engine);
+                            _ ->
+                                ct:fail(fail_to_unload_engine)
                         end;
+
                     {error, bad_engine_id} ->
                         {skip, "Dynamic Engine not supported"}
                 end
            catch
                error:notsup ->
-                  {skip, "Engine not supported on this SSL version"}
+                   {skip, "Engine not supported on this SSL version"}
            end
     end.
 
+%%-------------------------------------------------------------------------
+gc_clean()->
+    [{doc, "Test the special ensure load function."}].
+
+gc_clean(Config) when is_list(Config) ->
+    case crypto:get_test_engine() of
+        {error, notexist} ->
+            {skip, "OTP Test engine not found"};
+        {ok, Engine} ->
+
+	    Md5Hash1 = <<106,30,3,246,166,222,229,158,244,217,241,179,50,232,107,109>>,
+	    Md5Hash1 = crypto:hash(md5, "Don't panic"),
+	    Md5Hash2 =  <<0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15>>,
+
+	    load_without_unload(Engine, Md5Hash1, Md5Hash2),
+
+	    erlang:garbage_collect(),
+	    timer:sleep(1000),
+
+	    case crypto:hash(md5, "Don't panic") of
+		Md5Hash2 ->
+		    ct:fail(fail_to_unload_still_test_engine);
+		Md5Hash1 ->
+		    ok;
+		_ ->
+		    ct:fail(fail_to_unload_engine)
+	    end,
+	    
+	    try
+		case crypto:ensure_engine_loaded(<<"MD5">>, Engine) of
+		    {ok, E} ->
+			ok = crypto:engine_register(E, [engine_method_digests]),
+			case crypto:hash(md5, "Don't panic") of
+			    Md5Hash1 ->
+				ct:fail(fail_to_load_still_original_engine);
+			    Md5Hash2 ->
+				ok;
+			    _ ->
+				ct:fail(fail_to_load_engine)
+			end,
+			
+			ok = crypto:engine_unregister(E, [engine_method_digests]),
+			ok = crypto:engine_remove(E),
+			ok = crypto:ensure_engine_unloaded(E),
+			
+			case crypto:hash(md5, "Don't panic") of
+			    Md5Hash2 ->
+				ct:fail(fail_to_unload_still_test_engine);
+			    Md5Hash1 ->
+				ok;
+			    _ ->
+				ct:fail(fail_to_unload_engine)
+			end;
+		    {error, bad_engine_id} ->
+                        {skip, "Dynamic Engine not supported"}
+		end
+	     catch
+		 error:notsup ->
+			     {skip, "Engine not supported on this SSL version"}
+	     end
+    end.
+    
+load_without_unload(Engine, Md5Hash1, Md5Hash2) ->
+    case crypto:ensure_engine_loaded(<<"MD5">>, Engine) of
+	{ok, E} ->
+	    ok = crypto:engine_register(E, [engine_method_digests]),
+	    case crypto:hash(md5, "Don't panic") of
+		Md5Hash1 ->
+		    ct:fail(fail_to_load_still_original_engine);
+		Md5Hash2 ->
+		    ok = crypto:engine_unregister(E, [engine_method_digests]),
+		    ok;
+		_ ->
+		    ct:fail(fail_to_load_engine)
+	    end;
+	 {error, bad_engine_id} ->
+	    io:format("load_without_unload: bad_engine_id\n", [])
+    end.
+
 %%%----------------------------------------------------------------
-%%% Pub/priv key storage tests.  Thoose are for testing the crypto.erl
+%%% Pub/priv key storage tests.  Those are for testing the crypto.erl
 %%% support for using priv/pub keys stored in an engine.
 
 sign_verify_rsa(Config) ->
@@ -851,19 +943,27 @@ get_pub_from_priv_key_ecdsa(Config) ->
 %%%================================================================
 %%% Help for engine_stored_pub_priv_keys* test cases
 %%%
-load_storage_engine(Config) ->
-    load_storage_engine(Config, []).
+skip_if_unsup(Type, Config) ->
+    case pkey_supported(Type) of
+        false ->
+            {skip, "Unsupported in this cryptolib"};
+        true ->
+            Config
+    end.
 
-load_storage_engine(_Config, ExcludeMthds) ->
+
+pkey_supported(Type) ->
+    lists:member(Type, proplists:get_value(public_keys, crypto:supports(), [])).
+
+
+load_storage_engine(_Config) ->
     case crypto:get_test_engine() of
-        {ok, Engine} ->
+        {ok, EngineLibPath} ->
             try crypto:engine_load(<<"dynamic">>,
-                                   [{<<"SO_PATH">>, Engine},
-                                    <<"LOAD">>],
-                                   [],
-                                   crypto:engine_get_all_methods() -- ExcludeMthds
-                                  )
-            catch
+				   [{<<"SO_PATH">>, EngineLibPath},
+				    <<"LOAD">>],
+				   [])
+	    catch
                 error:notsup ->
                     {error, notsup}
             end;
