@@ -32,7 +32,8 @@
          io_with_huge_message_queue/1, format_string/1,
 	 maps/1, coverage/1, otp_14178_unicode_atoms/1, otp_14175/1,
          otp_14285/1, limit_term/1, otp_14983/1, otp_15103/1, otp_15076/1,
-         otp_15159/1, otp_15639/1]).
+         otp_15159/1, otp_15639/1, otp_15705/1, otp_15847/1, otp_15875/1,
+         chars_limit/1, otp_17525/1]).
 
 -export([pretty/2, trf/3]).
 
@@ -65,7 +66,7 @@ all() ->
      io_lib_width_too_small, io_with_huge_message_queue,
      format_string, maps, coverage, otp_14178_unicode_atoms, otp_14175,
      otp_14285, limit_term, otp_14983, otp_15103, otp_15076, otp_15159,
-     otp_15639].
+     otp_15639, otp_15705, otp_15847, otp_15875, chars_limit, otp_17525].
 
 %% Error cases for output.
 error_1(Config) when is_list(Config) ->
@@ -2504,9 +2505,11 @@ otp_14983(_Config) ->
 
 trunc_string() ->
     "str " = trf("str ", [], 10),
-    "str ..." = trf("str ~s", ["str"], 6),
+    "str str" = trf("str ~s", ["str"], 6),
+    "str ..." = trf("str ~s", ["str1"], 6),
     "str str" = trf("str ~s", ["str"], 7),
-    "str ..." = trf("str ~8s", ["str"], 6),
+    "str str" = trf("str ~8s", ["str"], 6),
+    "str ..." = trf("str ~8s", ["str1"], 6),
     Pa = filename:dirname(code:which(?MODULE)),
     {ok, UNode} = test_server:start_node(printable_range_unicode, slave,
 					 [{args, " +pc unicode -pa " ++ Pa}]),
@@ -2543,7 +2546,7 @@ trunc_depth(D, Fun) ->
     "#{{...} => {...},...}" = Fun(M, D, 7),
     "#{{[...],...} => {[...],...},...}" = Fun(M, D, 22),
     "#{{[...],...} => {[...],...},[...] => [...]}" = Fun(M, D, 31),
-    "#{{[...],...} => {[...],...},[...] => [...]}" = Fun(M, D, 33),
+    "#{{[...],...} => {[...],...},[1|...] => [...]}" = Fun(M, D, 33),
     "#{{[1|...],[...]} => {[1|...],[...]},[1,2|...] => [...]}" =
         Fun(M, D, 50),
 
@@ -2680,3 +2683,96 @@ otp_15639(_Config) ->
     "\"12345678\"..." = pretty("123456789"++[x], UOpts),
     "[[...]|...]" = pretty(["1","2","3","4","5","6","7","8"], UOpts),
     ok.
+
+otp_15705(_Config) ->
+    L = [<<"an">>,["at"],[["om"]]],
+    "..." = trf("~s", [L], 0),
+    "..." = trf("~s", [L], 1),
+    "..." = trf("~s", [L], 2),
+    "..." = trf("~s", [L], 3),
+    "a..." = trf("~s", [L], 4),
+    "an..." = trf("~s", [L], 5),
+    "anatom" = trf("~s", [L], 6),
+    L2 = ["a",[<<"na">>],[["tom"]]],
+    "..." = trf("~s", [L2], 3),
+    "a..." = trf("~s", [L2], 4),
+    "an..." = trf("~s", [L2], 5),
+    "anatom" = trf("~s", [L2], 6),
+
+    A = [[<<"äpple"/utf8>>, "plus", <<"äpple">>]],
+    "äp..." = trf("~ts", [A], 5),
+    "äppleplusäpple" = trf("~ts", [A], 14),
+    U = [["ки"],"рилл","и́ческий атом"],
+    "ки..." = trf("~ts", [U], 5),
+    "кирилли́ческий..." = trf("~ts", [U], 16),
+    "кирилли́ческий атом" = trf("~ts", [U], 20),
+
+    "|кирилли́чес|" = trf("|~10ts|", [U], -1),
+    ok.
+
+otp_15847(_Config) ->
+    T = {someRecord,<<"1234567890">>,some,more},
+    "{someRecord,<<...>>,...}" =
+        pretty(T, [{chars_limit,20}, {encoding,latin1}]),
+    ok.
+
+otp_15875(_Config) ->
+    %% This test is moot due to the fix in GH-4842.
+    S = io_lib:format("~tp", [[{0, [<<"00">>]}]], [{chars_limit, 18}]),
+    "[{0,[<<\"00\">>]}]" = lists:flatten(S).
+
+%% GH-4824, GH-4842, OTP-17459.
+chars_limit(_Config) ->
+    List = fun R(I) ->
+                   case I =:= 0 of true -> 0; false -> [I, R(I-1)] end
+           end,
+    Tuple = fun R(I) ->
+                   case I =:= 0 of true -> 0; false -> {I, R(I-1)} end
+           end,
+    Map = fun R(I) ->
+                   case I =:= 0 of true -> 0; false -> #{I => R(I-1)} end
+           end,
+    Record = fun R(I) ->
+                   case I =:= 0 of true -> 0; false -> {b, R(I-1)} end
+           end,
+    Test = fun (F, N, Lim) ->
+                   Opts = [{chars_limit, Lim},
+                           {record_print_fun, fun rfd/2}],
+                   [_|_] = io_lib_pretty:print(F(N), Opts)
+           end,
+    %% Used to loop:
+    Test(List, 1000, 1000),
+    Test(Tuple, 1000, 1000),
+    Test(Map, 1000, 1000),
+    Test(Record, 1000, 1000),
+
+    %% Misc sizes and char limits:
+    _ = [Test(What, N, CL) ||
+            N <- lists:seq(1, 50),
+            CL <- lists:seq(N, N*3),
+            What <- [List, Tuple, Map, Record]
+        ],
+    ok.
+
+%% GH-5053. 'chars_limit' bug.
+otp_17525(_Config) ->
+    L = [{xxxxxxxxx,aaaa},
+         {yyyyyyyyyyyy,1},
+         {eeeeeeeeeeeee,bbbb},
+         {ddddddddd,1111111111},
+         {gggggggggggggggggggg,cccc},
+         {uuuuuuuuuuuu,11}],
+    S = io_lib:format("aaaaaaaaaaaaaaaaaa ~p bbbbbbbbbbb ~p",
+                      ["cccccccccccccccccccccccccccccccccccccc", L],
+                      [{chars_limit, 155}]),
+    "aaaaaaaaaaaaaaaaaa \"cccccccccccccccccccccccccccccccccccccc\" bbbbbbbbbbb [{xxxxxxxxx,\n"
+    "                                                                          aaaa},\n"
+    "                                                                         {yyyyyyyyyyyy,\n"
+    "                                                                          1},\n"
+    "                                                                         {eeeeeeeeeeeee,\n"
+    "                                                                          bbbb},\n"
+    "                                                                         {ddddddddd,\n"
+    "                                                                          1111111111},\n"
+    "                                                                         {...}|...]" =
+    lists:flatten(S),
+    ok.
